"""
Document Store on SQLite.
"""

from __future__ import annotations

__all__ = ["SQLite"]

import json
import sqlite3
from typing import Any

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    ConflictError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    JSONHelper,
    ParameterParser,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
    Update,
    UpdateOp,
    Value,
)

from .._helper import (
    build_item_from_parts,
    build_item_from_value,
    build_query_result,
    get_collection_config,
)
from .._models import DocumentCollectionConfig, DocumentFieldType


class SQLite(StoreProvider):
    collection: str | None
    database: str
    table: str | None
    id_column: str | dict
    value_column: str | dict
    pk_column: str | dict | None
    id_map_field: str | dict | None
    pk_map_field: str | dict | None
    etag_embed_field: str | dict | None
    suppress_fields: list[str] | None
    nparams: dict[str, Any]

    _client: Any
    _aclient: Any
    _collection_cache: dict[str, SQLiteCollection]
    _acollection_cache: dict[str, SQLiteCollection]

    def __init__(
        self,
        database: str = ":memory:",
        table: str | None = None,
        id_column: str | dict = "id",
        value_column: str | dict = "value",
        pk_column: str | dict | None = "pk",
        id_map_field: str | dict | None = "id",
        pk_map_field: str | dict | None = "pk",
        etag_embed_field: str | dict | None = "_etag",
        suppress_fields: list[str] | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            database:
                SQLite database, defaults to ":memory:".
            table:
                SQLite table name mapped to document store collection.
            id_column:
                Column name to store the document id, defaults to "id".
            value_column:
                Column name to store the document store value.
                This column should have the JSON or JSONB type,
                defaults to "value".
            pk_column:
                Column name to store the document store pk,
                defaults to "pk".
            id_map_field:
                Field in the document to map into id.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "id".
            pk_map_field:
                Field in the document to map into pk.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "pk".
            etag_embed_field:
                Field to store the generated ETAG value.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the field, defaults to "_etag".
            suppress_fields:
                List of fields to supress when results are returned.
            nparams:
                Native parameters to sqlite client.
        """
        self.collection = table
        self.database = database
        self.table = table
        self.id_column = id_column
        self.value_column = value_column
        self.pk_column = pk_column
        self.id_map_field = id_map_field
        self.pk_map_field = pk_map_field
        self.etag_embed_field = etag_embed_field
        self.suppress_fields = suppress_fields
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._collection_cache = dict()
        self._acollection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        self._client = sqlite3.connect(
            self.database,
            check_same_thread=False,
            **self.nparams,
        )

    def _get_table_name(self, op_parser: StoreOperationParser) -> str:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        table = (
            collection_name if collection_name is not None else self.collection
        )
        if table is None:
            raise BadRequestError("Collection name must be specified.")
        return table

    def _get_collections(
        self, op_parser: StoreOperationParser
    ) -> list[SQLiteCollection]:
        if op_parser.is_resource_op():
            return []
        if op_parser.op_equals(StoreOperation.TRANSACT):
            collections: list[SQLiteCollection] = []
            for single_op_parser in op_parser.get_operation_parsers():
                collections.extend(self._get_collections(single_op_parser))
            return collections
        table = self._get_table_name(op_parser)
        if table is None:
            raise BadRequestError("Collection name must be specified")
        if table in self._collection_cache:
            return [self._collection_cache[table]]
        id_column = ParameterParser.get_collection_parameter(
            self.id_column, table
        )
        value_column = ParameterParser.get_collection_parameter(
            self.value_column, table
        )
        pk_column = ParameterParser.get_collection_parameter(
            self.pk_column, table
        )
        id_map_field = ParameterParser.get_collection_parameter(
            self.id_map_field, table
        )
        pk_map_field = ParameterParser.get_collection_parameter(
            self.pk_map_field, table
        )
        etag_embed_field = ParameterParser.get_collection_parameter(
            self.etag_embed_field, table
        )
        col = SQLiteCollection(
            table,
            id_column,
            value_column,
            pk_column,
            id_map_field,
            pk_map_field,
            etag_embed_field,
            self.suppress_fields,
        )
        self._collection_cache[table] = col
        return [col]

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[StoreOperation.PUT, StoreOperation.DELETE],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collections = self._get_collections(op_parser)
        ncall, state = self._get_ncall(
            op_parser, collections, ClientHelper(self._client)
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
            collections,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collections: list[SQLiteCollection],
        helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if len(collections) == 1:
            converter = collections[0].converter
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            table = self._get_table_name(op_parser)
            id_column = ParameterParser.get_collection_parameter(
                self.id_column, table
            )
            value_column = ParameterParser.get_collection_parameter(
                self.value_column, table
            )
            pk_column = ParameterParser.get_collection_parameter(
                self.pk_column, table
            )
            args = OperationConverter.convert_create_collection(
                table,
                get_collection_config(op_parser),
                {
                    "id_column": id_column,
                    "value_column": value_column,
                    "pk_column": pk_column,
                },
            )
            call = NCall(helper.execute, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = OperationConverter.convert_drop_collection(
                self._get_table_name(op_parser)
            )
            call = NCall(helper.execute, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = OperationConverter.convert_list_collections()
            call = NCall(helper.execute, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_key())
            call = NCall(helper.execute, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, state = converter.convert_put(
                op_parser.get_key(),
                op_parser.get_value(),
                op_parser.get_where(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: PreconditionFailedError},
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args, state = converter.convert_update(
                op_parser.get_key(),
                op_parser.get_set(),
                op_parser.get_where(),
                op_parser.get_returning(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: PreconditionFailedError},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_key(), op_parser.get_where()
            )
            call = NCall(helper.execute, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(helper.execute, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(helper.execute, args)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, state = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            call = NCall(helper.batch, args, None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            args, state = OperationConverter.convert_transact(
                op_parser.get_operation_parsers(),
                [col.converter for col in collections],
            )
            call = NCall(helper.transact, args, None)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collections: list[SQLiteCollection],
    ) -> Any:
        if len(collections) == 1:
            processor = collections[0].processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = []
            for item in nresult:
                result.append(item[0])
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if nresult is None:
                raise NotFoundError
            result = build_item_from_value(
                processor=processor,
                value=json.loads(nresult[0]),
                include_value=True,
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            where = op_parser.get_where()
            exists = op_parser.get_where_exists()
            if exists is False:
                pass
            elif where is not None:
                if nresult == 0:
                    raise PreconditionFailedError
            result = build_item_from_value(
                processor=processor,
                value=state["value"] if state is not None else None,
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            where = op_parser.get_where()
            returning = op_parser.get_returning()
            if where is None:
                if returning is None:
                    if nresult == 0:
                        raise NotFoundError
                else:
                    if nresult is None or len(nresult) == 0:
                        raise NotFoundError
            elif where is not None:
                if returning is None:
                    if nresult == 0:
                        raise PreconditionFailedError
                else:
                    if nresult is None or len(nresult) == 0:
                        raise PreconditionFailedError
            if isinstance(nresult, tuple):
                result = build_item_from_value(
                    processor=processor,
                    value=json.loads(nresult[0]),
                    include_value=True,
                )
            else:
                key = op_parser.get_key()
                result = build_item_from_parts(
                    processor=processor,
                    key=key,
                    etag=(
                        state["etag"]
                        if state is not None and "etag" in state
                        else None
                    ),
                )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            where = op_parser.get_where()
            if where is None:
                if nresult == 0:
                    raise NotFoundError
            elif where is not None:
                if nresult == 0:
                    raise PreconditionFailedError
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items: list = []
            select = op_parser.get_select()
            for item in nresult:
                if select is None or len(select.terms) == 0:
                    nvalue = json.loads(item[0])
                else:
                    nvalue = Helper.normalize_select(processor, select, item)
                items.append(
                    build_item_from_value(
                        processor=processor, value=nvalue, include_value=True
                    )
                )
            result = build_query_result(items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult[0]
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            if state is not None:
                for c in state["values"]:
                    if c is not None:
                        result.append(
                            build_item_from_value(
                                processor=processor,
                                value=c["value"],
                            )
                        )
                    else:
                        result.append(None)
        # TRANSACT
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            op_parsers = op_parser.get_operation_parsers()
            for i in range(0, len(op_parsers)):
                if state is not None:
                    c = state["values"][i]
                result.append(
                    self._convert_nresult(
                        nresult[i], c, op_parsers[i], [collections[i]]
                    )
                )
        return result


class Helper:
    @staticmethod
    def normalize_select(
        processor: ItemProcessor, select: Select, item: tuple
    ) -> dict:
        result: dict = {}
        for i in range(0, len(select.terms)):
            JSONHelper.update_field(
                result,
                processor.resolve_field(select.terms[i].field),
                UpdateOp.PUT,
                json.loads(item[i]),
            )
        return result


class OperationConverter:
    processor: ItemProcessor
    table: str
    id_column: str
    value_column: str
    pk_column: str | None

    FIELD_TYPE_TEXT: str = "TEXT"
    FIELD_TYPE_NUMERIC: str = "NUMERIC"
    FIELD_TYPE_BOOLEAN: str = "BOOLEAN"
    FIELD_TYPE_JSONB: str = "JSONB"

    def __init__(
        self,
        processor: ItemProcessor,
        table: str,
        id_column: str,
        value_column: str,
        pk_column: str | None,
    ):
        self.processor = processor
        self.table = table
        self.id_column = id_column
        self.value_column = value_column
        self.pk_column = pk_column

    @staticmethod
    def convert_create_collection(
        table: str, config: DocumentCollectionConfig | None, columns: dict
    ) -> dict:
        def get_column_type(type: str) -> str:
            if type == DocumentFieldType.STRING:
                return "TEXT"
            elif type == DocumentFieldType.NUMBER:
                return "NUMERIC"
            raise BadRequestError("Field type not supported")

        pk_column = (
            config.pk_field
            if config is not None and config.pk_field is not None
            else columns["pk_column"]
        )
        pk_type = get_column_type(
            config.pk_type
            if config is not None and config.pk_type is not None
            else DocumentFieldType.STRING
        )
        id_column = (
            config.id_field
            if config is not None and config.id_field is not None
            else columns["id_column"]
        )
        id_type = get_column_type(
            config.id_type
            if config is not None and config.id_type is not None
            else DocumentFieldType.STRING
        )
        value_column = (
            config.value_field
            if config is not None and config.value_field is not None
            else columns["value_column"]
        )

        if pk_column is not None:
            str = f"""CREATE TABLE IF NOT EXISTS {table}
                ({id_column} {id_type} PRIMARY KEY,
                {pk_column} {pk_type},
                {value_column} JSON,
                UNIQUE({id_column}))
                """
        else:
            str = f"""CREATE TABLE IF NOT EXISTS {table}
                ({id_column} {id_type} PRIMARY KEY,
                {value_column} JSON,
                UNIQUE({id_column}))
                """
        return {"query": str}

    @staticmethod
    def convert_drop_collection(table: str) -> dict:
        str = f"DROP TABLE {table}"
        return {"query": str}

    @staticmethod
    def convert_list_collections() -> dict:
        str = "SELECT name FROM sqlite_master WHERE type='table'"
        return {"query": str, "fetchall": True}

    @staticmethod
    def convert_transact(
        op_parsers: list[StoreOperationParser],
        converters: list[OperationConverter],
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for i in range(0, len(op_parsers)):
            op_parser = op_parsers[i]
            converter = converters[i]
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = converter.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.UPDATE):
                args, state = converter.convert_update(
                    op_parser.get_key(),
                    op_parser.get_set(),
                    op_parser.get_where(),
                    op_parser.get_returning(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = converter.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append(args)
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, dict]:
        ops: list = []
        states: list = []
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.PUT):
                args, state = self.convert_put(
                    op_parser.get_key(),
                    op_parser.get_value(),
                    op_parser.get_where(),
                    op_parser.get_where_exists(),
                )
                ops.append(args)
                states.append(state)
            elif op_parser.op_equals(StoreOperation.DELETE):
                args = self.convert_delete(
                    op_parser.get_key(), op_parser.get_where()
                )
                ops.append(args)
                states.append(None)
        return {"ops": ops}, {"values": states}

    def convert_get(self, key: Value) -> dict:
        str = f"""SELECT {self.value_column} FROM {self.table}
                    WHERE {self.convert_key_comparison(key)}"""
        return {"query": str}

    def convert_put(
        self,
        key: Value,
        value: dict,
        where: Expression | None,
        exists: bool | None,
    ) -> tuple[dict, dict | None]:
        document = self.processor.add_embed_fields(value, key)
        if key is not None:
            key = self.processor.get_key_from_key(key)
        else:
            key = self.processor.get_key_from_value(document)
        if (where is None and exists is None) or exists is False:
            id, pk = self.processor.get_id_pk_from_key(key)
            if self.pk_column is not None:
                columns = f"""({self.id_column},
                    {self.pk_column}, {self.value_column})"""
                values = f"""({self.convert_expr(id)},
                    {self.convert_expr(pk)}, '{json.dumps(document)}')"""
            else:
                columns = f"({self.id_column}, {self.value_column})"
                values = f"({self.convert_expr(id)}, '{json.dumps(document)}')"
            query = f"INSERT INTO {self.table} {columns} VALUES {values}"
            if exists is None:
                query = f"""{query} ON CONFLICT ({self.id_column})
                    DO UPDATE SET {self.value_column} =
                    EXCLUDED.{self.value_column}"""
        elif exists is True or where is not None:
            where_expr = self.convert_key_comparison(key)
            if exists is not True and where is not None:
                where_expr = f"{where_expr} AND {self.convert_expr(where)}"
            query = f"""UPDATE {self.table}
                SET {self.value_column} = '{json.dumps(document)}'
                WHERE {where_expr}"""
        return {"query": query, "rowcount": True}, {"value": document}

    def convert_update(
        self,
        key: Value,
        set: Update,
        where: Expression | None,
        returning: bool | None,
    ) -> tuple[dict, dict | None]:
        state = None
        if self.processor.needs_local_etag():
            etag = self.processor.generate_etag()
            set = self.processor.add_etag_update(set, etag)
            state = {"etag": etag}
        str = f"UPDATE {self.table} SET {self.convert_update_ops(set)}"
        where_expr = self.convert_key_comparison(key)
        if where is not None:
            where_expr = f"{where_expr} AND {self.convert_expr(where)}"
        str = f"{str} WHERE {where_expr}"
        if returning:
            str = f"{str} RETURNING {self.value_column}"
        if not returning:
            return {"query": str, "rowcount": True}, state
        return {"query": str}, state

    def convert_delete(self, key: Value, where: Expression | None) -> dict:
        where_expr = self.convert_key_comparison(key)
        if where is not None:
            where_expr = f"{where_expr} AND {self.convert_expr(where)}"
        str = f"DELETE FROM {self.table} WHERE {where_expr}"
        return {"query": str, "rowcount": True}

    def convert_query(
        self,
        select: Select | None = None,
        where: Expression | None = None,
        order_by: OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict:
        str = f"""SELECT {self.convert_select(
            select)} FROM {self.table}"""
        if where is not None:
            str = f"{str} WHERE {self.convert_expr(where)}"
        if order_by is not None:
            str = f"""{str} ORDER BY {self.convert_order_by(
                order_by)}"""
        if limit is not None:
            if offset is not None:
                str = f"""{str} LIMIT {limit} OFFSET {offset}"""
            else:
                str = f"""{str} LIMIT {limit}"""
        return {"query": str, "fetchall": True}

    def convert_count(self, where: Expression | None = None) -> dict:
        str = f"SELECT COUNT(*) FROM {self.table}"
        if where is not None:
            str = f"{str} WHERE {self.convert_expr(where)}"
        return {"query": str}

    def convert_key_comparison(self, key: Value) -> str:
        id, pk = self.processor.get_id_pk_from_key(key)
        expr = f"{self.id_column} = {self.convert_expr(id)}"
        if self.pk_column is not None:
            expr = f"{expr} AND {self.pk_column} = {self.convert_expr(pk)}"
        return expr

    def convert_field(
        self, field: Field | str, field_type: str | None = None
    ) -> str:
        path: str
        if isinstance(field, str):
            path = self.processor.resolve_field(field)
        elif isinstance(field, Field):
            path = self.processor.resolve_field(field.path)
        path = (
            path.replace(".", "/")
            .replace("[", "/")
            .replace("]", "")
            .rstrip("/")
        )
        splits = path.split("/")
        nsplits = []
        for split in splits:
            if split.isnumeric():
                nsplits.append(split)
            else:
                nsplits.append(f"'{split}'")
        splits = nsplits
        splits.insert(0, self.value_column)
        if (
            field_type == OperationConverter.FIELD_TYPE_TEXT
            or field_type == OperationConverter.FIELD_TYPE_NUMERIC
            or field_type == OperationConverter.FIELD_TYPE_BOOLEAN
        ):
            path = "->".join(splits[0:-1])
            path = f"{path}->>{splits[-1]}"
        else:
            path = "->".join(splits)
        return path

    def convert_expr(self, expr: Expression | None) -> str:
        if expr is None:
            return "NULL"
        if isinstance(expr, str):
            return f"'{expr}'"
        if isinstance(expr, bool):
            return str(expr).upper()
        if isinstance(expr, (int, float)):
            return json.dumps(expr)
        if isinstance(expr, (dict, list)):
            return f"'{json.dumps(expr, separators=(',', ':'))}'"
        if isinstance(expr, Field):
            return self.convert_field(expr)
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            return self.convert_comparison(expr)
        if isinstance(expr, And):
            return f"""({self.convert_expr(
                expr.lexpr)} AND {self.convert_expr(
                expr.rexpr)})"""
        if isinstance(expr, Or):
            return f"""({self.convert_expr(
                expr.lexpr)} OR {self.convert_expr(
                expr.rexpr)})"""
        if isinstance(expr, Not):
            return f"""NOT {self.convert_expr(
                expr.expr)}"""
        return str(expr)

    def convert_comparison(self, expr: Comparison) -> str:
        def get_field_type(value: Any) -> str | None:
            ftype = None
            if isinstance(value, bool):
                ftype = OperationConverter.FIELD_TYPE_BOOLEAN
            elif isinstance(value, (int, float)):
                ftype = OperationConverter.FIELD_TYPE_NUMERIC
            elif isinstance(value, str):
                ftype = OperationConverter.FIELD_TYPE_TEXT
            elif isinstance(value, list) and len(value) > 0:
                return get_field_type(value[0])
            return ftype

        value = None
        if isinstance(expr.lexpr, Field):
            value = expr.rexpr
        elif isinstance(expr.rexpr, Field):
            value = expr.lexpr
        field_type = get_field_type(value)

        if isinstance(expr.lexpr, Field):
            lhs = self.convert_field(expr.lexpr, field_type)
        else:
            lhs = self.convert_expr(expr.lexpr)
        if isinstance(expr.rexpr, Field):
            rhs = self.convert_field(expr.rexpr, field_type)
        else:
            rhs = self.convert_expr(expr.rexpr)

        if expr.op == ComparisonOp.BETWEEN and isinstance(expr.rexpr, list):
            return f"""{lhs} >= {self.convert_expr(
                expr.rexpr[0])} AND {lhs} <= {self.convert_expr(
                expr.rexpr[1])}"""
        elif (
            expr.op == ComparisonOp.IN or expr.op == ComparisonOp.NIN
        ) and isinstance(expr.rexpr, list):
            lst = ", ".join(self.convert_expr(i) for i in expr.rexpr)
            rhs = f"({lst})"
        return f"""{lhs} {expr.op.value} {rhs}"""

    def convert_func(self, expr: Function) -> str:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_TYPE:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                field_type = DocumentFieldType(args[1].lower())
                if field_type == DocumentFieldType.OBJECT:
                    field_types = ["'object'"]
                elif field_type == DocumentFieldType.ARRAY:
                    field_types = ["'array'"]
                elif field_type == DocumentFieldType.BOOLEAN:
                    field_types = ["'false'", "'true'"]
                elif field_type == DocumentFieldType.NUMBER:
                    field_types = ["'integer'", "'real'"]
                elif field_type == DocumentFieldType.STRING:
                    field_types = ["'text'"]
                elif field_type == DocumentFieldType.NULL:
                    field_types = ["'null'"]
                in_arr = f"({','.join(field_types)})"
                return f"JSON_TYPE({self.value_column}, '{field}') IN {in_arr}"
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_field(args[0])
                return f"{field} IS NOT NULL"
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_field(args[0])
                return f"{field} IS NULL"
            if name == StoreFunctionName.LENGTH:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                return f"LENGTH({field})"
            if name == StoreFunctionName.CONTAINS:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                value = args[1]
                return f"{field} LIKE '%{value}%'"
            if name == StoreFunctionName.STARTS_WITH:
                field = self.convert_field(
                    args[0], OperationConverter.FIELD_TYPE_TEXT
                )
                value = args[1]
                return f"{field} LIKE '{value}%'"
            if name == StoreFunctionName.ARRAY_LENGTH:
                field = self.convert_field(args[0])
                return f"JSON_ARRAY_LENGTH({field})"
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                value = self.convert_expr(args[1])
                where = f"value = {value}"
                json_each = (
                    f"JSON_EACH({self.table}.{self.value_column}, '{field}')"
                )
                query = f"SELECT 1 FROM {json_each} WHERE {where}"
                query = f"EXISTS ({query})"
                return query
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.processor.resolve_field(args[0])
                field = f"$.{field}"
                clauses = []
                for item in args[1]:
                    value = self.convert_expr(item)
                    clauses.append(f"value = {value}")
                where = f"({str.join(' OR ', clauses)})"
                json_each = (
                    f"JSON_EACH({self.table}.{self.value_column}, '{field}')"
                )
                query = f"SELECT 1 FROM {json_each} WHERE {where}"
                query = f"EXISTS ({query})"
                return query
        raise BadRequestError(f"Function {name} not supported")

    def convert_order_by(self, order_by: OrderBy) -> str:
        str_terms = []
        for term in order_by.terms:
            _str_term = self.convert_field(term.field)
            if term.direction is not None:
                _str_term = f"{_str_term} {term.direction.value}"
            str_terms.append(_str_term)
        return ", ".join([t for t in str_terms])

    def convert_select(self, select: Select | None) -> str:
        if select is None or len(select.terms) == 0:
            return self.value_column
        str_terms = []
        for term in select.terms:
            _str_term = self.convert_expr(Field(path=term.field))
            if term.alias is not None:
                _str_term = f"{_str_term} AS {term.alias}"
            str_terms.append(_str_term)
        return ", ".join([t for t in str_terms])

    def convert_update_ops(self, update: Update) -> str:
        def get_field_path(field: str) -> tuple:
            field = self.processor.resolve_field(field)
            path = (
                field.replace(".", "/")
                .replace("[", "/")
                .replace("]", "")
                .rstrip("/")
            )
            splits = path.split("/")
            if field.endswith("[-]"):
                field = field.replace("[-]", "[#]")
            field = f"'$.{field}'"
            return field, splits

        str = self.value_column
        for operation in update.operations:
            op = operation.op
            field_path, splits = get_field_path(operation.field)
            if op == UpdateOp.PUT:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0], (dict, list)):
                    value = f"json({value})"
                str = f"JSON_SET({str}, {field_path}, {value})"
            elif op == UpdateOp.INSERT:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0], (dict, list)):
                    value = f"json({value})"
                if splits[-1] == "-":
                    str = f"JSON_INSERT({str}, {field_path}, {value})"
                elif splits[-1].isnumeric():
                    index = int(splits[-1])
                    array_field = f"{field_path[: field_path.rfind('[')]}'"
                    if isinstance(operation.args[0], (dict, list)):
                        group = "JSON_GROUP_ARRAY(json(t5.value))"
                    else:
                        group = "JSON_GROUP_ARRAY(t5.value)"
                    value_query = f"""(SELECT {group}
                        FROM (SELECT t2.value
                        FROM (SELECT t1.value
                        FROM JSON_EACH({self.table}.{self.value_column},
                        {array_field})
                        AS t1 LIMIT {index})
                        AS t2
                        UNION ALL
                        SELECT {value}
                        UNION ALL
                        SELECT t4.value
                        FROM (SELECT t3.value
                        FROM JSON_EACH({self.table}.{self.value_column},
                        {array_field})
                        AS t3 LIMIT -1 OFFSET {index})
                        AS t4)
                        AS t5)"""
                    str = f"JSON_SET({str}, {array_field}, {value_query})"
                else:
                    str = f"JSON_SET({str}, {field_path}, {value})"
            elif op == UpdateOp.DELETE:
                str = f"JSON_REMOVE({str}, {field_path})"
            elif op == UpdateOp.INCREMENT:
                value = operation.args[0]
                extract = f"JSON_EXTRACT({self.value_column}, {field_path})"
                extract = f"IFNULL({extract}, 0)"
                extract = f"{extract} + {value}"
                str = f"JSON_SET({str}, {field_path}, {extract})"
            elif op == UpdateOp.MOVE:
                dest_field_path, s = get_field_path(operation.args[0].path)
                extract = f"JSON_EXTRACT({self.value_column}, {field_path})"
                str = f"JSON_SET({str}, {dest_field_path}, {extract})"
                str = f"JSON_REMOVE({str}, {field_path})"
            elif op == UpdateOp.ARRAY_UNION:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0][0], (dict, list)):
                    group = "JSON_GROUP_ARRAY(json(t3.value))"
                else:
                    group = "JSON_GROUP_ARRAY(t3.value)"
                value_query = f"""(SELECT {group}
                    FROM (SELECT t1.value
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    AS t1
                    UNION ALL
                    SELECT t2.value
                    FROM JSON_EACH({value})
                    AS t2
                    WHERE NOT EXISTS (
                    SELECT 1
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    WHERE value = t2.value))
                    AS t3)
                    """
                str = f"JSON_SET({str}, {field_path}, {value_query})"
            elif op == UpdateOp.ARRAY_REMOVE:
                value = self.convert_expr(operation.args[0])
                if isinstance(operation.args[0][0], (dict, list)):
                    group = "JSON_GROUP_ARRAY(json(t2.value))"
                else:
                    group = "JSON_GROUP_ARRAY(t2.value)"
                value_query = f"""(SELECT {group}
                    FROM (SELECT t1.value
                    FROM JSON_EACH({self.table}.{self.value_column},
                    {field_path})
                    AS t1
                    WHERE NOT EXISTS (
                    SELECT 1
                    FROM JSON_EACH({value})
                    WHERE value = t1.value))
                    AS t2)
                    """
                str = f"JSON_SET({str}, {field_path}, {value_query})"
            else:
                raise BadRequestError(
                    f"Operation {operation.op.value} is not supported"
                )
        return f"{self.value_column} = {str}"


class ClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    def execute(
        self, query: str, rowcount: bool = False, fetchall: bool = False
    ) -> Any:
        cursor = self.client.cursor()
        try:
            cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return cursor.fetchall()
            return cursor.fetchone()
        finally:
            self.client.commit()
            cursor.close()

    def batch(self, ops: list) -> Any:
        cursor = self.client.cursor()
        try:
            for op in ops:
                cursor.execute(op["query"])
        finally:
            self.client.commit()
            cursor.close()

    def transact(self, ops: list) -> Any:
        result = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    if cursor.rowcount == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(cursor.rowcount)
                else:
                    res = cursor.fetchone()
                    if res is None or len(res) == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(res)
            self.client.commit()
        finally:
            cursor.close()
        return result

    def close(self, nargs: Any) -> Any:
        pass


class AsyncClientHelper:
    client: Any

    def __init__(self, client: Any):
        self.client = client

    async def execute(
        self, query: str, rowcount: bool = False, fetchall: bool = False
    ) -> Any:
        cursor = self.client.cursor()
        try:
            await cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return await cursor.fetchall()
            return await cursor.fetchone()
        finally:
            await self.client.commit()
            await cursor.close()

    async def close(self, nargs: Any) -> Any:
        await self.client.close()


class SQLiteCollection:
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(
        self,
        table: str,
        id_column: str,
        value_column: str,
        pk_column: str | None,
        id_map_field: str | None,
        pk_map_field: str | None,
        etag_embed_field: str | None,
        suppress_fields: list[str] | None,
    ) -> None:
        self.processor = ItemProcessor(
            etag_embed_field=etag_embed_field,
            id_map_field=id_map_field,
            pk_map_field=pk_map_field,
            local_etag=True,
            suppress_fields=suppress_fields,
        )
        self.converter = OperationConverter(
            self.processor,
            table,
            id_column,
            value_column,
            pk_column,
        )



"""
Config Store on Azure App Configuration.
"""

__all__ = ["AzureAppConfiguration"]

from typing import Any

from azure.appconfiguration import ConfigurationSetting
from azure.core.exceptions import ResourceNotFoundError

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import QueryArgs, get_query_args, normalize_label
from .._models import ConfigItem, ConfigKey, ConfigList, ConfigProperties


class AzureAppConfiguration(StoreProvider):
    connection_string: str | None
    base_url: str | None
    credential_type: str | None
    tenant_id: str | None
    client_id: str | None
    client_secret: str | None
    certificate_path: str | None
    nparams: dict[str, Any]

    _credential: Any
    _acredential: Any
    _client: Any
    _aclient: Any
    _processor: ItemProcessor

    def __init__(
        self,
        connection_string: str | None = None,
        base_url: str | None = None,
        credential_type: str | None = None,
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            connection_string:
                Azure App Configuration connection string.
                If connection string is provided,
                base_url is optional and ignored.
            base_url:
                Azure App Configuration base url.
                If connection_string is not provided, base_url is required.
            credential_type:
                Azure credential type.
                Credential types are default, client_secret, certificate,
                azure_cli, shared_token_cache, managed_identity.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Azure App Configuration client.
        """
        self.connection_string = connection_string
        self.base_url = base_url
        self.credential_type = credential_type
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.certificate_path = certificate_path
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from azure.appconfiguration import AzureAppConfigurationClient

        from verse.internal.storage_common._azure import get_credential

        self._credential, self._client = self._init_credential_client(
            AzureAppConfigurationClient,
            get_credential,
        )

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from azure.appconfiguration.aio import AzureAppConfigurationClient

        from verse.internal.storage_common._azure import aget_credential

        self._acredential, self._aclient = self._init_credential_client(
            AzureAppConfigurationClient,
            aget_credential,
        )

    def _init_credential_client(
        self,
        app_config_client,
        get_credential,
    ):
        credential = None
        if self.base_url is not None:
            credential = get_credential(
                credential_type=self.credential_type,
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.client_secret,
                certificate_path=self.certificate_path,
            )
            client = app_config_client(
                base_url=self.base_url,
                credential=credential,
                **self.nparams,
            )
        else:
            client = app_config_client.from_connection_string(
                self.connection_string, **self.nparams
            )
        return credential, client

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client, self._credential),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._aclient,
            AsyncClientHelper(self._aclient, self._acredential),
        )
        if ncall is None:
            return await super().arun(operation)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()

        # GET
        if op_parser.op_equals(StoreOperation.GET):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            args: dict[str, Any] = {"key": id, "label": label}
            call = NCall(
                client.get_configuration_setting,
                args,
                nargs,
                {ResourceNotFoundError: NotFoundError},
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            label = normalize_label(op_parser.get_label())
            config = ConfigurationSetting(
                key=id,
                label=label,
                value=value,
            )
            args = {"configuration_setting": config}
            call = NCall(client.set_configuration_setting, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            label = normalize_label(op_parser.get_label())
            args = {"key": id, "label": label}
            call = NCall(
                client.delete_configuration_setting,
                args,
                nargs,
                {None: NotFoundError},
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = {
                "args": OperationConverter.convert_filter(
                    get_query_args(op_parser)
                ),
                "nargs": nargs,
            }
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = {
                "args": OperationConverter.convert_filter(
                    get_query_args(op_parser)
                ),
                "nargs": nargs,
            }
            call = NCall(helper.query, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None

        def convert_item(item):
            return ConfigItem(
                key=ConfigKey(id=item.key, label=item.label),
                value=item.value,
                properties=ConfigProperties(updated_time=item.last_modified),
            )

        # GET
        if op_parser.op_equals(StoreOperation.GET):
            result = convert_item(nresult)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = convert_item(nresult)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult:
                items.append(convert_item(item))
            items = sorted(items, key=lambda x: (x.key.label, x.key.id))
            result = ConfigList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            count = 0
            for item in nresult:
                count = count + 1
            result = count
        return result


class OperationConverter:
    @staticmethod
    def convert_filter(query_args: QueryArgs) -> dict:
        args = {}
        if query_args.id_filter is not None:
            args["key_filter"] = f"{query_args.id_filter}*"
        if query_args.label_filter is not None:
            args["label_filter"] = query_args.label_filter
        return args


class ClientHelper:
    client: Any
    credential: Any

    def __init__(self, client: Any, credential: Any):
        self.client = client
        self.credential = credential

    def query(self, args: dict, nargs: Any):
        return NCall(
            self.client.list_configuration_settings,
            args,
            nargs,
        ).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class AsyncClientHelper:
    client: Any
    credential: Any

    def __init__(self, client: Any, credential: Any):
        self.client = client
        self.credential = credential

    async def query(self, args: dict, nargs: Any):
        response = NCall(
            self.client.list_configuration_settings,
            args,
            nargs,
        ).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def close(self, nargs: Any):
        await NCall(
            self.client.close,
            None,
            nargs,
        ).ainvoke()
        if self.credential is not None:
            await NCall(
                self.credential.close,
                None,
                nargs,
            ).ainvoke()



"""
Secret Store on HashiCorp Vault.
"""

__all__ = ["HashiCorpVault"]

from datetime import datetime
from typing import Any

import hvac
from hvac.exceptions import InvalidPath, InvalidRequest

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    NotFoundError,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)


class HashiCorpVault(StoreProvider):
    url: str
    token: str | None
    role_id: str | None
    secret_id: str | None
    namespace: str | None
    value_field: str
    nparams: dict[str, Any]

    _client: Any
    _processor: ItemProcessor

    def __init__(
        self,
        url: str,
        token: str | None = None,
        role_id: str | None = None,
        secret_id: str | None = None,
        namespace: str | None = "admin",
        value_field: str = "value",
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            url:
                Vault url.
            token:
                Vault token.
            role_id:
                Vault role id.
            secret_id:
                Vault secret id.
            namespace:
                Vault namespace. Defaults to "admin".
            value_field:
                Value field name to store the secret value.
                Defaults to "value".
            nparams:
                Native parameters to vault client.
        """
        self.url = url
        self.token = token
        self.role_id = role_id
        self.secret_id = secret_id
        self.namespace = namespace
        self.value_field = value_field
        self.nparams = nparams

        self._client = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        if self.token is not None:
            self._client = hvac.Client(
                url=self.url,
                namespace=self.namespace,
                token=self.token,
                **self.nparams,
            )
        elif self.role_id is not None or self.secret_id is not None:
            self._client = hvac.Client(
                url=self.url,
                namespace=self.namespace,
                **self.nparams,
            )
            auth_response = self._client.auth.approle.login(
                role_id=self.role_id,
                secret_id=self.secret_id,
            )
            token = auth_response["auth"]["client_token"]
            self._client.token = token

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(
                self._client,
                self.value_field,
            ),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            args: dict[str, Any] = {
                "path": id,
                "version": version,
                "raise_on_deleted_version": True,
            }
            call = NCall(
                client.secrets.kv.read_secret_version,
                args,
                nargs,
                {InvalidPath: NotFoundError},
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {
                "path": id,
                "raise_on_deleted_version": True,
            }
            call = NCall(
                client.secrets.kv.read_secret,
                args,
                nargs,
                {InvalidPath: NotFoundError},
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"path": id}
            call = NCall(
                client.secrets.kv.v2.read_secret_metadata,
                args,
                nargs,
                {InvalidPath: NotFoundError},
            )
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata = op_parser.get_metadata()
            exists = op_parser.get_where_exists()
            args = {
                "path": id,
                "secret": dict({self.value_field: value}),
            }
            if metadata is None and exists is None:
                call = NCall(
                    client.secrets.kv.v2.create_or_update_secret,
                    args,
                    nargs,
                )
            elif metadata is None and exists is False:
                args["cas"] = 0
                call = NCall(
                    client.secrets.kv.v2.create_or_update_secret,
                    args,
                    nargs,
                    {InvalidRequest: PreconditionFailedError},
                )
            else:
                args = {
                    "id": id,
                    "value": value,
                    "metadata": metadata if metadata is not None else {},
                    "where_exists": exists,
                    "nargs": nargs,
                }
                call = NCall(helper.put, args)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            args = {"id": id, "value": value, "nargs": nargs}
            call = NCall(
                helper.update,
                args,
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            args = {
                "id": id,
                "metadata": metadata if metadata is not None else {},
                "nargs": nargs,
            }
            call = NCall(
                helper.update_metadata,
                args,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            args = {"id": id, "nargs": nargs}
            call = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = {"path": ""}
            call = NCall(
                client.secrets.kv.v2.list_secrets,
                args,
                nargs,
                {InvalidPath: None},
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = {"path": ""}
            call = NCall(
                client.secrets.kv.v2.list_secrets,
                args,
                nargs,
                {InvalidPath: None},
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(
                    id=id, version=str(nresult["data"]["metadata"]["version"])
                ),
                value=(
                    nresult["data"]["data"][self.value_field]
                    if self.value_field in nresult["data"]["data"]
                    else None
                ),
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            metadata: dict | None = {}
            if nresult["data"]["metadata"]["custom_metadata"] is not None:
                metadata = nresult["data"]["metadata"]["custom_metadata"]
            result = SecretItem(
                key=SecretKey(id=id),
                metadata=metadata if metadata is not None else {},
                properties=SecretProperties(
                    created_time=Helper.convert_to_datetime(
                        nresult["data"]["metadata"]["created_time"]
                    )
                ),
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            versions: list = []
            for key, value in nresult["data"]["versions"].items():
                versions.append(
                    SecretVersion(
                        version=str(key),
                        created_time=Helper.convert_to_datetime(
                            value["created_time"]
                        ),
                    )
                )
            versions.sort(key=lambda s: int(s.version), reverse=True)
            return SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id, version=str(nresult["data"]["version"]))
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            result = SecretItem(
                key=SecretKey(id=id, version=str(nresult["data"]["version"]))
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            result = SecretItem(
                key=SecretKey(id=id),
                metadata=metadata if metadata is not None else {},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            if nresult is not None:
                for id in nresult["data"]["keys"]:
                    items.append(SecretItem(key=SecretKey(id=id)))
            items = QueryProcessor.query_items(
                items=items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_root_field,
            )
            result = SecretList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = []
            if nresult is not None:
                for id in nresult["data"]["keys"]:
                    items.append(SecretItem(key=SecretKey(id=id)))
            result = QueryProcessor.count_items(
                items=items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_root_field,
            )
        return result


class Helper:
    @staticmethod
    def convert_to_datetime(timestamp: str):
        return datetime.strptime(
            timestamp[:-4],
            "%Y-%m-%dT%H:%M:%S.%f",
        ).astimezone()


class ClientHelper:
    client: Any
    value_field: str

    def __init__(self, client: Any, value_field: str):
        self.client = client
        self.value_field = value_field

    def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool | None,
        nargs: Any,
    ) -> Any:
        if where_exists is True:
            NCall(
                self.client.secrets.kv.read_secret,
                {"path": id, "raise_on_deleted_version": True},
                nargs,
                {InvalidPath: PreconditionFailedError},
            ).invoke()
        if metadata is not None:
            response = NCall(
                self.client.secrets.kv.v2.create_or_update_secret,
                (
                    {
                        "path": id,
                        "secret": dict({self.value_field: value}),
                        "cas": 0,
                    }
                    if where_exists is False
                    else {
                        "path": id,
                        "secret": dict({self.value_field: value}),
                    }
                ),
                nargs,
                (
                    {InvalidRequest: PreconditionFailedError}
                    if where_exists is False
                    else None
                ),
            ).invoke()
            NCall(
                self.client.secrets.kv.v2.update_metadata,
                {"path": id, "custom_metadata": metadata},
                nargs,
            ).invoke()
            return response
        raise BadRequestError("Operation not supported")

    def update(self, id: str, value: str, nargs: Any) -> Any:
        NCall(
            self.client.secrets.kv.read_secret,
            {"path": id, "raise_on_deleted_version": True},
            nargs,
            {InvalidPath: NotFoundError},
        ).invoke()
        return NCall(
            self.client.secrets.kv.v2.create_or_update_secret,
            {"path": id, "secret": dict({self.value_field: value})},
            nargs,
        ).invoke()

    def update_metadata(self, id: str, metadata: dict, nargs: Any) -> Any:
        NCall(
            self.client.secrets.kv.read_secret,
            {"path": id, "raise_on_deleted_version": True},
            nargs,
            {InvalidPath: NotFoundError},
        ).invoke()
        return NCall(
            self.client.secrets.kv.v2.update_metadata,
            {"path": id, "custom_metadata": metadata},
            nargs,
        ).invoke()

    def delete(self, id: str, nargs: Any) -> Any:
        NCall(
            self.client.secrets.kv.read_secret,
            {"path": id, "raise_on_deleted_version": True},
            nargs,
            {InvalidPath: NotFoundError},
        ).invoke()
        return NCall(
            self.client.secrets.kv.v2.delete_metadata_and_all_versions,
            {"path": id},
            nargs,
        ).invoke()

    def close(self, nargs: Any) -> Any:
        pass



from __future__ import annotations

from typing import Any

from verse.core import Operation, Response
from verse.internal.storage_core import (
    StoreComponent,
    StoreOperation,
    UpdateAttribute,
)
from verse.ql import Expression, OrderBy, Select, Update

from ._models import (
    VectorCollectionConfig,
    VectorItem,
    VectorKey,
    VectorList,
    VectorValue,
)


class VectorStore(StoreComponent):
    def __init__(self, **kwargs):
        super().__init__(
            **kwargs,
        )

    def create_collection(
        self,
        collection: str | None = None,
        config: dict | VectorCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.

        Returns:
            None.
        """
        operation = StoreOperation.create_collection(
            collection=collection,
            config=config,
            **kwargs,
        )
        return self.run(operation)

    def drop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop collection.

        Args:
            collection:
                Collection name.

        Returns:
            None.
        """
        operation = StoreOperation.drop_collection(
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def list_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        operation = StoreOperation.list_collections(
            **kwargs,
        )
        return self.run(operation)

    def get(
        self,
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Get vector.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Vector item.

        Raises:
            NotFoundError:
                Vector not found.
        """
        operation = StoreOperation.get(
            key=key,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def put(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        metadata: dict | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Put vector.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            metadata:
                Custom metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        operation = StoreOperation.put(
            key=key,
            value=value,
            metadata=metadata,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def update(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector value.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        update = Update()
        update.put(UpdateAttribute.VALUE, value)
        operation = StoreOperation.update(
            key=key,
            set=update,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def update_metadata(
        self,
        key: str | dict | VectorKey,
        metadata: dict | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector metadata.

        Args:
            key:
                Vector key.
            metadata:
                Vector metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        update = Update()
        update.put(UpdateAttribute.METADATA, metadata)
        operation = StoreOperation.update(
            key=key,
            set=update,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def delete(
        self,
        key: str | dict | VectorKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete vector.

        Args:
            key:
                Vector key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Vector not found.
            PreconditionFailedError:
                Condition failed.
        """
        operation = StoreOperation.delete(
            key=key,
            where=where,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def query(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[VectorList]:
        """Query vectors.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            List of vectors.
        """
        operation = StoreOperation.query(
            select=select,
            search=search,
            where=where,
            order_by=order_by,
            limit=limit,
            offset=offset,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def count(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count vectors.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Vector count.
        """
        operation = StoreOperation.count(
            search=search,
            where=where,
            collection=collection,
            **kwargs,
        )
        return self.run(operation)

    def batch(
        self,
        operations: list[Operation],
        **kwargs,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            operations:
                List of operations.

        Returns:
            Batch operation results.
        """
        operation = StoreOperation.batch(
            operations=operations,
            **kwargs,
        )
        return self.run(operation)

    def transact(
        self,
        operations: list[Operation],
        **kwargs,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            operations:
                List of operations.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        operation = StoreOperation.transact(
            operations=operations,
            **kwargs,
        )
        return self.run(operation)

    def close(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the sync client.

        Returns:
            None.
        """
        operation = StoreOperation.close(
            **kwargs,
        )
        return self.run(operation)

    async def acreate_collection(
        self,
        collection: str | None = None,
        config: dict | VectorCollectionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create collection.

        Args:
            collection:
                Collection name.
            config:
                Collection config.

        Returns:
            None.
        """
        operation = StoreOperation.create_collection(
            collection=collection,
            config=config,
            **kwargs,
        )
        return await self.arun(operation)

    async def adrop_collection(
        self,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop collection.

        Args:
            collection:
                Collection name.

        Returns:
            None.
        """
        operation = StoreOperation.drop_collection(
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def alist_collections(
        self,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List collections.

        Returns:
            List of collection names.
        """
        operation = StoreOperation.list_collections(
            **kwargs,
        )
        return await self.arun(operation)

    async def aget(
        self,
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Get vector.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Vector item.

        Raises:
            NotFoundError:
                Vector not found.
        """
        operation = StoreOperation.get(
            key=key,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def aput(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        metadata: dict | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Put vector.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            metadata:
                Custom metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        operation = StoreOperation.put(
            key=key,
            value=value,
            metadata=metadata,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def aupdate(
        self,
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector value.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        update = Update()
        update.put(UpdateAttribute.VALUE, value)
        operation = StoreOperation.update(
            key=key,
            set=update,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def aupdate_metadata(
        self,
        key: str | dict | VectorKey,
        metadata: dict | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[VectorItem]:
        """Update vector metadata.

        Args:
            key:
                Vector key.
            metadata:
                Vector metadata.
            collection:
                Collection name.

        Returns:
            Vector item.
        """
        update = Update()
        update.put(UpdateAttribute.METADATA, metadata)
        operation = StoreOperation.update(
            key=key,
            set=update,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def adelete(
        self,
        key: str | dict | VectorKey | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Delete vector.

        Args:
            key:
                Vector key.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            None.

        Raises:
            NotFoundError:
                Vector not found.
            PreconditionFailedError:
                Condition failed.
        """
        operation = StoreOperation.delete(
            key=key,
            where=where,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def aquery(
        self,
        search: str | Expression | None = None,
        select: str | Select | None = None,
        where: str | Expression | None = None,
        order_by: str | OrderBy | None = None,
        limit: int | None = None,
        offset: int | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[VectorList]:
        """Query vectors.

        Args:
            search:
                Search expression.
            select:
                Select expression.
            where:
                Condition expression.
            order_by:
                Order by expression.
            limit:
                Query limit.
            offset:
                Query offset.
            collection:
                Collection name.

        Returns:
            List of vectors.
        """
        operation = StoreOperation.query(
            select=select,
            search=search,
            where=where,
            order_by=order_by,
            limit=limit,
            offset=offset,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def acount(
        self,
        search: str | Expression | None = None,
        where: str | Expression | None = None,
        collection: str | None = None,
        **kwargs,
    ) -> Response[int]:
        """Count vectors.

        Args:
            search:
                Search expression.
            where:
                Condition expression.
            collection:
                Collection name.

        Returns:
            Vector count.
        """
        operation = StoreOperation.count(
            search=search,
            where=where,
            collection=collection,
            **kwargs,
        )
        return await self.arun(operation)

    async def abatch(
        self,
        operations: list[Operation],
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Batch operation.

        Args:
            operations:
                List of operations.

        Returns:
            Batch operation results.
        """
        operation = StoreOperation.batch(
            operations=operations,
            **kwargs,
        )
        return await self.arun(operation)

    async def atransact(
        self,
        operations: list[Operation],
        **kwargs: Any,
    ) -> Response[list[Any]]:
        """Execute transaction.

        Args:
            operations:
                List of operations.

        Returns:
            Transaction results.

        Raises:
            ConflictError:
                Transaction failed.
        """
        operation = StoreOperation.transact(
            operations=operations,
            **kwargs,
        )
        return await self.arun(operation)

    async def aclose(
        self,
        **kwargs: Any,
    ) -> Response[None]:
        """Close the async client.

        Returns:
            None.
        """
        operation = StoreOperation.close(
            **kwargs,
        )
        return await self.arun(operation)


class VectorOperation:
    @staticmethod
    def get(
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Get vector operation.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Get vector operation.
        """
        return StoreOperation.get(
            key=key,
            collection=collection,
            **kwargs,
        )

    @staticmethod
    def put(
        key: str | dict | VectorKey,
        value: list[float] | dict | VectorValue,
        metadata: dict | None,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Any:
        """Put vector operation.

        Args:
            key:
                Vector key.
            value:
                Vector value.
            metadata:
                Custom metadata.
            collection:
                Collection name.

        Returns:
            Put vector operation.
        """
        return StoreOperation.put(
            key=key,
            value=value,
            metadata=metadata,
            collection=collection,
            **kwargs,
        )

    @staticmethod
    def delete(
        key: str | dict | VectorKey,
        collection: str | None = None,
        **kwargs: Any,
    ) -> Any:
        """Delete vector operation.

        Args:
            key:
                Vector key.
            collection:
                Collection name.

        Returns:
            Delete vector operation.
        """
        return StoreOperation.delete(
            key=key,
            collection=collection,
            **kwargs,
        )



from verse.core import DataModel


class KeyValueKey(DataModel):
    """Key Value key."""

    id: str
    """Key Value id.
    """


class KeyValueItem(DataModel):
    """Config item."""

    key: KeyValueKey
    """Key Value key.
    """

    value: str | bytes | None = None
    """Config value.
    """


class KeyValueList(DataModel):
    """Key Value item list."""

    items: list[KeyValueItem]
    """List of Key Value items.
    """



"""
Secret Store on ENV file.
"""

__all__ = ["EnvFile"]

from typing import Any

from dotenv import dotenv_values

from verse.core import Context, Operation, Response
from verse.core.exceptions import NotFoundError
from verse.internal.storage_core import StoreOperation, StoreProvider

from .._constants import LATEST_VERSION
from .._models import SecretItem, SecretKey


class EnvFile(StoreProvider):
    path: str = ".env"

    _items: dict | None

    def __init__(self, path: str = ".env", **kwargs):
        """Initalize.

        Args:
            path:
                Path to file. Defaults to ".env".
        """
        self.path = path
        self._items = None

    def init(self, context: Context | None = None) -> None:
        if self._items is not None:
            return

        self._items = dotenv_values(self.path)
        if self._items is None:
            self._items = dict()

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        result = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            if self._items is None:
                raise NotFoundError
            if id not in self._items:
                raise NotFoundError
            val = self._items[id]
            result = SecretItem(
                key=SecretKey(id=id, version=LATEST_VERSION), value=val
            )
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            pass
        else:
            return super().run(operation)
        return Response(result=result)



"""
Secret Store on Azure Key Vault.
"""

__all__ = ["AzureKeyVault"]

from typing import Any

from azure.core.exceptions import ResourceNotFoundError

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import NotFoundError, PreconditionFailedError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._models import (
    SecretItem,
    SecretKey,
    SecretList,
    SecretProperties,
    SecretVersion,
)


class AzureKeyVault(StoreProvider):
    vault_url: str
    credential_type: str | None
    tenant_id: str | None
    client_id: str | None
    client_secret: str | None
    certificate_path: str | None
    nparams: dict[str, Any]

    _credential: Any
    _acredential: Any
    _client: Any
    _aclient: Any
    _processor: ItemProcessor

    def __init__(
        self,
        vault_url: str,
        credential_type: str | None = "default",
        tenant_id: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        certificate_path: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs
    ):
        """Initialize.

        Args:
            vault_url:
                Azure Key Vault url.
            credential_type:
                Azure credential type.
                Credential types are default, client_secret, certificate,
                azure_cli, shared_token_cache, managed_identity.
            tenant_id:
                Azure tenant id for client_secret credential type.
            client_id:
                Azure client id for client_secret credential type.
            client_secret:
                Azure client secret for client_secret credential type.
            certificate_path:
                Certificate path for certificate credential type.
            nparams:
                Native parameters to Azure Key Vault client.
        """
        self.vault_url = vault_url
        self.credential_type = credential_type
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.certificate_path = certificate_path
        self.nparams = nparams

        self._client = None
        self._aclient = None
        self._processor = ItemProcessor()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from azure.keyvault.secrets import SecretClient

        from verse.internal.storage_common._azure import get_credential

        self._credential, self._client = self._init_credential_client(
            SecretClient, get_credential
        )

    async def ainit(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from azure.keyvault.secrets.aio import SecretClient

        from verse.internal.storage_common._azure import aget_credential

        self._acredential, self._aclient = self._init_credential_client(
            SecretClient,
            aget_credential,
        )

    def _init_credential_client(
        self,
        secret_client,
        get_credential,
    ):
        credential = get_credential(
            credential_type=self.credential_type,
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            certificate_path=self.certificate_path,
        )
        client = secret_client(vault_url=self.vault_url, credential=credential)
        return credential, client

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._client,
            ClientHelper(self._client, self._credential),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    async def arun(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        await self.ainit(context=context)
        op_parser = self.get_op_parser(operation)
        ncall = self._get_ncall(
            op_parser,
            self._aclient,
            AsyncClientHelper(self._aclient, self._acredential),
        )
        if ncall is None:
            return await super().arun(operation)
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(nresult, op_parser, self._processor)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        client: Any,
        helper: Any,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            args: dict[str, Any] = {"name": id, "version": version}
            call = NCall(
                client.get_secret,
                args,
                nargs,
                {ResourceNotFoundError: NotFoundError},
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"name": id}
            call = NCall(
                client.get_secret,
                args,
                nargs,
                {ResourceNotFoundError: NotFoundError},
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            id = op_parser.get_id_as_str()
            args = {"id": id, "nargs": nargs}
            call = NCall(helper.get_versions, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            id = op_parser.get_id_as_str()
            value = op_parser.get_value()
            metadata = op_parser.get_metadata()
            exists = op_parser.get_where_exists()
            if exists is None:
                args = {
                    "name": id,
                    "value": value,
                    "tags": metadata if metadata is not None else {},
                }
                call = NCall(
                    client.set_secret,
                    args,
                    nargs,
                    {ResourceNotFoundError: NotFoundError},
                )
            elif exists is False or exists is True:
                args = {
                    "id": id,
                    "value": value,
                    "metadata": metadata if metadata is not None else {},
                    "where_exists": exists,
                    "nargs": nargs,
                }
                call = NCall(helper.put, args)
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            id = op_parser.get_id_as_str()
            value = op_parser.get_updated_value()
            args = {"id": id, "value": value, "nargs": nargs}
            call = NCall(helper.update, args)
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            id = op_parser.get_id_as_str()
            metadata = op_parser.get_updated_metadata()
            args = {
                "name": id,
                "tags": metadata if metadata is not None else {},
            }
            call = NCall(
                client.update_secret_properties,
                args,
                nargs,
                {ResourceNotFoundError: NotFoundError},
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            id = op_parser.get_id_as_str()
            args = {"id": id, "nargs": nargs}
            call = NCall(helper.delete, args)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = {"nargs": nargs}
            call = NCall(helper.query, args)
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(helper.close, args)
        else:
            return None
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: StoreOperationParser,
        processor: ItemProcessor,
    ) -> Any:
        result: Any = None
        # GET value
        if (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            result = SecretItem(
                key=SecretKey(
                    id=nresult.name, version=nresult.properties.version
                ),
                value=nresult.value,
            )
        # GET metadata
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_metadata_attribute()
        ):
            result = SecretItem(
                key=SecretKey(id=nresult.name),
                metadata=nresult.properties.tags,
                properties=SecretProperties(
                    created_time=nresult.properties.created_on
                ),
            )
        # GET versions
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_versions_attribute()
        ):
            versions: list = []
            for item in nresult:
                id = item.name
                versions.append(
                    SecretVersion(
                        version=item.version,
                        created_time=item.created_on,
                    )
                )
            if len(versions) == 0:
                raise NotFoundError
            versions = sorted(
                versions,
                key=lambda s: s.created_time,
                reverse=True,
            )
            return SecretItem(key=SecretKey(id=id), versions=versions)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = SecretItem(
                key=SecretKey(
                    id=nresult.name, version=nresult.properties.version
                )
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = SecretItem(
                key=SecretKey(
                    id=nresult.name, version=nresult.properties.version
                )
            )
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = SecretItem(
                key=SecretKey(id=nresult.name),
                metadata=nresult.tags,
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.name),
                        metadata=item.tags,
                        properties=SecretProperties(
                            created_time=item.created_on
                        ),
                    )
                )
            items = QueryProcessor.query_items(
                items=items,
                select=op_parser.get_select(),
                where=op_parser.get_where(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
                field_resolver=processor.resolve_root_field,
            )
            result = SecretList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            items = []
            for item in nresult:
                items.append(
                    SecretItem(
                        key=SecretKey(id=item.name),
                        metadata=item.tags,
                        properties=SecretProperties(
                            created_time=item.created_on
                        ),
                    )
                )
            result = QueryProcessor.count_items(
                items=items,
                where=op_parser.get_where(),
                field_resolver=processor.resolve_root_field,
            )
        return result


class ClientHelper:
    client: Any
    credential: Any

    def __init__(self, client: Any, credential: Any):
        self.client = client
        self.credential = credential

    def get_versions(self, id: str, nargs: Any):
        return NCall(
            self.client.list_properties_of_secret_versions,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).invoke()

    def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool,
        nargs: Any,
    ) -> Any:
        response = NCall(
            self.client.get_secret,
            {"name": id},
            nargs,
            {
                ResourceNotFoundError: (
                    None if not where_exists else PreconditionFailedError
                ),
            },
        ).invoke()
        if not where_exists and response is not None:
            raise PreconditionFailedError
        return NCall(
            self.client.set_secret,
            {"name": id, "value": value, "tags": metadata},
            nargs,
        ).invoke()

    def update(self, id: str, value: str, nargs: Any) -> Any:
        response = NCall(
            self.client.get_secret,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).invoke()
        return NCall(
            self.client.set_secret,
            {"name": id, "value": value, "tags": response.properties.tags},
            nargs,
        ).invoke()

    def delete(self, id: str, nargs: Any) -> Any:
        poller = NCall(
            self.client.begin_delete_secret,
            {"name": id},
            nargs,
            {
                # Secret could be in soft-delete state.
                # If secret doesn't exist, purge call will throw error.
                ResourceNotFoundError: None,
            },
        ).invoke()
        if poller is not None:
            poller.wait()
        NCall(
            self.client.purge_deleted_secret,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).invoke()

    def query(self, nargs: Any):
        return NCall(
            self.client.list_properties_of_secrets,
            None,
            nargs,
        ).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class AsyncClientHelper:
    client: Any
    credential: Any

    def __init__(self, client: Any, credential: Any):
        self.client = client
        self.credential = credential

    async def get_versions(self, id: str, nargs: Any):
        response = NCall(
            self.client.list_properties_of_secret_versions,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def put(
        self,
        id: str,
        value: str,
        metadata: dict | None,
        where_exists: bool,
        nargs: Any,
    ) -> Any:
        response = await NCall(
            self.client.get_secret,
            {"name": id},
            nargs,
            {
                ResourceNotFoundError: (
                    None if not where_exists else PreconditionFailedError
                ),
            },
        ).ainvoke()
        if not where_exists and response is not None:
            raise PreconditionFailedError
        return await NCall(
            self.client.set_secret,
            {"name": id, "value": value, "tags": metadata},
            nargs,
        ).ainvoke()

    async def update(self, id: str, value: str, nargs: Any) -> Any:
        response = await NCall(
            self.client.get_secret,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).ainvoke()
        return await NCall(
            self.client.set_secret,
            {"name": id, "value": value, "tags": response.properties.tags},
            nargs,
        ).ainvoke()

    async def delete(self, id: str, nargs: Any):
        await NCall(
            self.client.delete_secret,
            {"name": id},
            nargs,
            {
                # Secret could be in soft-delete state.
                # If secret doesn't exist, purge call will throw error.
                ResourceNotFoundError: None,
            },
        ).ainvoke()
        await NCall(
            self.client.purge_deleted_secret,
            {"name": id},
            nargs,
            {ResourceNotFoundError: NotFoundError},
        ).ainvoke()

    async def query(self, nargs: Any):
        response = NCall(
            self.client.list_properties_of_secrets,
            None,
            nargs,
        ).invoke()
        nresult = []
        async for item in response:
            nresult.append(item)
        return nresult

    async def close(self, nargs: Any):
        await NCall(
            self.client.close,
            None,
            nargs,
        ).ainvoke()
        await NCall(
            self.credential.close,
            None,
            nargs,
        ).ainvoke()



from verse.core.exceptions import BadRequestError
from verse.internal.storage_core import (
    KeyAttribute,
    SpecialAttribute,
    StoreOperationParser,
    StoreFunctionName,
)
from verse.ql import Comparison, ComparisonOp, Field, Function, FunctionNamespace

from ._constants import DEFAULT_LABEL


def normalize_label(label: str | None):
    return label if label is not None else DEFAULT_LABEL


class QueryArgs:
    id_filter: str | None = None
    label_filter: str | None = None


def get_query_args(
    op_parser: StoreOperationParser,
) -> QueryArgs:
    args = QueryArgs()
    expr_list = op_parser.get_where_expr_list()
    not_supported = False
    for expr in expr_list:
        if isinstance(expr, Function):
            if expr.namespace == FunctionNamespace.BUILTIN:
                if expr.name == StoreFunctionName.STARTS_WITH:
                    field = expr.args[0].path
                    prefix = expr.args[1]
                    if (
                        field == KeyAttribute.ID
                        or field == SpecialAttribute.ID
                    ):
                        args.id_filter = prefix
                    else:
                        not_supported = True
                else:
                    not_supported = True
        elif isinstance(expr, Comparison):
            if isinstance(expr.lexpr, Field) and (
                expr.rexpr is None or isinstance(expr.rexpr, str)
            ):
                field = expr.lexpr.path
                value = expr.rexpr
                if (
                    field == KeyAttribute.LABEL
                    or field == SpecialAttribute.LABEL
                ):
                    if expr.op == ComparisonOp.EQ:
                        args.label_filter = value
                    else:
                        not_supported = True
                else:
                    not_supported = True
            elif isinstance(expr.rexpr, Field) and (
                expr.lexpr is None or isinstance(expr.lexpr, str)
            ):
                field = expr.rexpr.path
                value = expr.lexpr
                if (
                    field == KeyAttribute.LABEL
                    or field == SpecialAttribute.LABEL
                ):
                    if expr.op == ComparisonOp.EQ:
                        args.label_filter = value
                    else:
                        not_supported = True
                else:
                    not_supported = True
        else:
            not_supported = True
        if not_supported:
            raise BadRequestError("Condition not supported")
    return args
